#!/usr/bin/perl
#
use WWW::Mechanize;
use DBI;
use DBD::mysql;
use Digest::MD5 qw(md5 md5_hex md5_base64);
use Time::Local;

# this package has been created ad-hoc to solve a bug found inside
# WWW::Mechanize. The subclass overloads _push_page_stack method
# avoiding the history stack growing bigger than 10

package WWW::Mechanize::KeepSlim;

our @ISA = qw/WWW::Mechanize/;

sub _push_page_stack {
	my $self = shift;

	if ( $self->{res} ) {
		my $save_stack = $self->{page_stack};
    		$self->{page_stack} = [];
    		push( @$save_stack, $self->clone );

        	# HERE! - stop the stack getting bigger than 10
    		if ( @$save_stack > 10 ) {
	    		shift(@$save_stack);
    		}
    		$self->{page_stack} = $save_stack;
	}
	return 1;
}

package main;

# DB stuff ------------------------------------------------------------------
my $db		= "deli";
my $host	= "localhost";
my $port	= "3306";
my $dsn		= "DBI:mysql:database=$db;host=$host;port=$port";
my $user	= "icious";
my $pass	= "blahblah";

my $dbh		= DBI->connect($dsn, $user, $pass, {
			RaiseError=>0,
			PrintError=>0,
			AutoCommit=>1,
			HandleError=>\&dbierrorlog}
		);
die "Could not connect to the DB\n\n" unless $dbh;

# global vars ---------------------------------------------------------------

# ----------- client variables
my $CLIENT	= `hostname`; 	# this is used by the distributed version
chomp $CLIENT;  		# take away the \n at the end of hostname output

my $MAX_URL_SIZE= 0; 		# if 0 => no limits for the URL (should not be changed)

my $LOGFILE	= "./data/deli.log";	# this is the name of the log file
my $DEBUG	= 2; 		# this variable accepts debug verbosity levels from 0 to 4
my $WARNINGS	= 1;		# this one turns on and off warnings

# ----------- site variables (these might change if delicious modifies its structure)
my $URL		= "http://del.icio.us/###USER###/?setcount=100";
# if the program breaks at a very high user page number, I sometimes hack the
# code like this for a single run which resumes from that page.
#my $URL		= "http://del.icio.us/###USER###/?setcount=100&page=56";
#

# ----------- connection variables (you can tweak them to get better performance/stability
my $SLEEP	= 5;		# if it's <4 seconds, sooner or later they might kick you out
my $BIGSLEEP	= 30;		# this is a longer sleep, used if a connection breaks 
my $RETRIES	= 3;		# raise this if the program hangs often due to connection probs
	
my $USE_PROXY	= 0;		# if yahoo bans you, set this to 1 ;-)
my $PROXY	= "http://localhost:8118"; # I use tor to anonymize my connections...

# ----------- program options
# ---------------------------------------------------------------------------
# ---------------------------------------------------------------------------

my %h;		# handler for queries... this method is quite horrible for now ;)
prepare_q (\%h);

my $mech	= WWW::Mechanize::KeepSlim->new();
my $mech2	= WWW::Mechanize::KeepSlim->new();

if ($USE_PROXY){
	$mech->proxy(['http', 'ftp'], $PROXY);
	$mech2->proxy(['http', 'ftp'], $PROXY);
}

my ($quid, $user);

# try to resume
($qid, $user) = resume ($CLIENT);

# get user from queue
($qid, $user) = get_user_from_q($dbh) unless $qid;

print "[ ] Queue is empty.\n" unless $qid;

while ($qid){

	print "[-] Got id: $qid, name: $user from queue\n" if $DEBUG>=2;
	print "[*] Working on user $user:\n";
	# got user: save it inside the DB
	my $uid = insert_user ($user, $h{u1}, $h{u2});
	
	my $url = $URL;
	$url =~ s/###USER###/$user/;
	my $pagenum = 1;	
	$mech->get($url);

	do{
		print "[-] Downloaded page $pagenum for user $user\n" if $DEBUG;
		sleep $SLEEP;
		
		my $posts = skipordie ($mech->content, $user, $pagenum);
		while ($posts =~/<li class="post".*?>(.*?)<\/li>/gsi){
	
			my ($url, $title, $date, @tags) = parse_content($1);

			# if post data is not extracted just skip (a warning will be logged)
			next unless $url;
			# if url is longer than MAXSIZE char skip it
			if ($MAX_URL_SIZE && length($url)>$MAX_URL_SIZE){
				logmessage ("URL from user $user at page $pagenum is too long. URL follows:\n$url");
				next;
			}

			# save resource
			print "[-] URL: $url\n" if $DEBUG;
			print "[-] DATE: $date\n" if $DEBUG>=2;
			print "[-] TITLE: $title\n" if $DEBUG>=2;
			my $rid = insert_res ($url, $title, $h{r1}, $h{r2}, \$new);
			print "[-] TAGS: ".join (" ", @tags). "\n" if $DEBUG>=2;

			# if @tags is empty add one default tag called system:unfiled
			push (@tags, "system:unfiled") unless @tags;
			
			foreach my $tag (@tags){
				my $tid = insert_tag ($tag, $h{t1}, $h{t2});
				insert_tagged ($uid, $tid, $rid, $date, $h{x1}, $h{x2});
			}
		}
		$pagenum++;
		print "[-] Downloading page $pagenum for user $user...\n" if $DEBUG;

	# NOTE: the regexp for the "earlier" link should be corrected to be more precise
	# 	(this way it follows every first link whose text ends with "earlier")
	} while ($mech->follow_link(text_regex => qr/ earlier$/, n => 1) != undef);

	# update user timestamp
	$h{u3}->execute(getdate(), $uid);

	# if everything went ok delete user from queue...
	$h{q3}->execute($qid);

	# and get next user in queue
	($qid, $user) = get_user_from_q($dbh);
}

# ---------------------------------------------------------------------------
# ---------------------------------------------------------------------------

sub skipordie{
	my ($content,$user, $pagenum) = @_;
	
	if ($content =~/<ol class="posts">(.*?)<\/ol>/si){
		return $1;
	}

	# this error is generated by privoxy filter, to solve it I just switch the
	# proxy off and then on after I load the page.
	if ($content =~ /<title>Request blocked/i){
		print "[w] Page blocked by privoxy. Switching the proxy off for a while...\n";
		logmessage ("[w] Page blocked by privoxy. Switching the proxy off and on.");
		
		$mech->proxy(['http', 'ftp'], '');
	
		print "[w] Loading page...\n";	
		$mech->reload();

		print "[w] Switching the proxy back on!\n";
		$mech->proxy(['http', 'ftp'], $PROXY);
			
		if ($mech->content =~/<ol class="posts">(.*?)<\/ol>/si){
			return $1;
		}

		logmessage ("Given up on user $user at page $pagenum.");
		die "[x] Error during proxy (de)activation: giving up\n";
		
	}

	if ($content =~ /page not found/i){
		print "[w] Page not found for user $user. I think I'll dump this on log file...\n";
		logmessage ("Page not found for user $user.");
		return;
	}

	my @retry_errors = (
				'500 read timeout', 
				'Connection: close', 
				"500 Can\'t connect", 
				'Internal Error',
				'500 Server closed connection without sending any data back',
				);
	foreach my $error (@retry_errors){
		if ($content =~ /$error/si){
			print "[w] Had an error ($error) while downloading the page. I'll try to download it again.\n";
			logmessage ("Error ($error) while downloading page $pagenum for user $user.");

			my $retries = $RETRIES;
			# note: this still could be fixed, 'cause I might enter this with an error
			# 	and then have another one when I reload the page. But we hope this
			# 	doesn't happen so often!
			while (($content =~ /$error/i)&&($retries)){
				print "[w] Attempt ". $RETRIES - $retries + 1 ."/$RETRIES... \n";
				sleep ($BIGSLEEP);
				
				# a normal reload works most of the times
				$mech->reload();

				# if reload doesn't work you may try to create a new mech
				#print "[w] Creating a new mech which connects to $url.&page=$pagenum\n";
				#$mech = WWW::Mechanize::KeepSlim->new();
				#$mech->get($url."&page=$pagenum");
				#$content = $mech->content;

				$retries--;
			}

			if (!$RETRIES){
				logmessage ("Given up on user $user at page $pagenum.");
				die "[x] Too many retries: giving up\n";
			}

			if ($mech->content =~/<ol class="posts">(.*?)<\/ol>/si){
				return $1;
			}
		}
	}
	logmessage ("Could not extract posts from user $user (page $pagenum) and don't know why. Content follows:\n".$mech->content);
	die "[x] Could not extract posts from user $user (page $pagenum) and don't know why. Content follows:\n".$mech->content;
}

# ---------------------------------------------------------------------------
sub logmessage{
	my $message = shift;
	my $date = getdate();
	
	open OUT, ">>$LOGFILE";
	print OUT "[$date] $message\n";
	close OUT;
}

# ---------------------------------------------------------------------------
sub get_user_from_q{
        my ($dbh) = @_;

        print "[-] Locking table 'userq'\n" if $DEBUG>=3;
        $dbh->do("LOCK TABLES userq as u1 WRITE, userq as u2 WRITE");
        
	# select first unlocked user

        $h{q1}->execute;
        my ($id, $name) = $h{q1}->fetchrow_array();

	# once you get the id, lock the user
	$h{q2}->execute($CLIENT, $id);

        print "[-] Unlocking table 'userq'\n" if $DEBUG>=3;
	$dbh->do("UNLOCK TABLES");

	return ($id, $name);
}

# ---------------------------------------------------------------------------

sub parse_content{
	my $post = shift; 
	my ($url, $title, $date, @tags);

	# note: the a.*?href is used to work both with 'a href' and 'a class="mp3" href'
	if ($post =~ /class="desc"><a.*?href="(.*?)".*?>(.*?)<\/a>.*?<div class="meta">(.*?)<\/div>/si){
		$url = $1;
		$title = $2;
		my $meta = $3;
		print "*** META BEGIN ***\n$meta\n*** META END ***\n" if $DEBUG>=3;
		while ($meta =~ /<a class="tag".*?>(.*?)<\/a>/gsi){
			push @tags, $1;
		}
		if ($meta =~ /span class="date" title="(.*?)"/si){
			$date = $1;
			chop ($date);
			$date =~ s/T/ /g;				
		}else{
			 die "[x] Could not extract date";
		}
	}else{
		print "[w] Could not extract post data. Meta follows:\n$meta\n";
		logmessage ("[w] Could not extract post data. Meta follows:\n$meta\n");
	}

	return ($url, $title, $date, @tags);
}


# ---------------------------------------------------------------------------
sub dbierrorlog{
	my $error = shift;
	print "$error\n" if $DEBUG>=3;
	logmessage ($error);
}

# ---------------------------------------------------------------------------
sub resume{
	my ($client) = @_;

	$h{q0}->execute($client);
	my ($id, $name) = $h{q0}->fetchrow_array();

	if ($id){
		print "[w] Found user $name (id: $id) locked by this client ($client). Resuming...\n";
	}
	return ($id, $name);
}

# ---------------------------------------------------------------------------
sub prepare_q{
	my ($h) = @_;

	$$h{u1}  = $dbh->prepare("select id from user where name = ?") || die ("Error in prepare");
	$$h{u2}  = $dbh->prepare("insert into user (name) values (?)") || die ("Error in prepare");
	$$h{u3}  = $dbh->prepare("update user set bkpdate=? where id=?") || die ("Error in prepare");
	$$h{r1}  = $dbh->prepare("select id from res where md5 = ?") || die ("Error in prepare");
	$$h{r2}  = $dbh->prepare("insert into res(md5, url, title) values (?,?,?)") || die ("Error in prepare");
	$$h{t1}  = $dbh->prepare("select id from tag where name = ?") || die ("Error in prepare");
	$$h{t2}  = $dbh->prepare("insert into tag(name) values (?)") || die ("Error in prepare");
	$$h{x1}  = $dbh->prepare("select uid from tagged where uid=? and tid=? and rid=?") || die ("Error in prepare");
	$$h{x2}  = $dbh->prepare("insert into tagged(uid, tid, rid, date) values (?,?,?,?)") || die ("Error in prepare");
	$$h{q0}  = $dbh->prepare("select id, name from userq where locked=1 and client=?") || die ("Error in prepare");
	$$h{q1}  = $dbh->prepare("select id, name from userq as u1 where id=(select min(id) from userq as u2 where locked=0)") || die ("Error in prepare");
        $$h{q2}  = $dbh->prepare("update userq as u1 set locked=1,client=? where id=?") || die ("Error in prepare");
        $$h{q3}  = $dbh->prepare("delete from userq where id=?") || die ("Error in prepare");
        $$h{q4}  = $dbh->prepare("insert into userq (name) values (?)") || die ("Error in prepare");

}

# ---------------------------------------------------------------------------
sub lock_table{
	my ($table) = @_;
        
	print "[-] Locking table $table\n" if $DEBUG>=4;
	$dbh->do("LOCK TABLES $table WRITE") || die ("Error executing lock on table $table");
}

# ---------------------------------------------------------------------------
sub unlock_table{
	my ($table) = @_;
	
	print "[-] Unlocking table $table\n" if $DEBUG>=4;
	$dbh->do("UNLOCK TABLES") || die ("Error unlocking tables");
}

# ---------------------------------------------------------------------------
sub insert_user{
	my ($user, $uh1, $uh2) = @_;

	#1) Lock the table (for writing)
        lock_table ("user");
	
	#2) Check if user exists inside the db
	$uh1->execute($user);
	my $id = ($uh1->fetchrow_array())[0];

	#3) If user does not exist insert it
	if (!$id){
		print "[w] User $user does not exist: creating...\n" if $WARNINGS;
		$uh2->execute($user) || die ("Error in execute");

		# NOTE: using last_insert_id to get auto_increment field value.
		# 	This usage (with 4 null values) is compatible with mysql
		# 	but could be uncompatible with other databases (yes,
		# 	I'm lazy...)
		$id = $dbh->last_insert_id(null,null,null,null); 

		# this is a longer way to get auto_increment field value
		# $uh1->execute($user);
		# $id = ($uh1->fetchrow_array())[0];
		
	}
	
	#4) Unlock the table
	unlock_table("user");
	
	return $id;
}

# ---------------------------------------------------------------------------
sub insert_tag{
	my ($tag, $th1, $th2) = @_;

	#1) Lock the table (for writing)
        lock_table ("tag");

	#2) Check if tag exists inside the db	
	$th1->execute($tag);
	my $id = ($th1->fetchrow_array())[0];

	#3) If the tag does not exist insert it
	if (!$id){
		print "[w] Tag $tag does not exist: creating...\n" if $WARNINGS;
		$th2->execute($tag) || die ("Error in execute");
		$id = $dbh->last_insert_id(null,null,null,null); 
	}

	#4) Unlock the table
	unlock_table("tag");
	
	return $id;
}

# ---------------------------------------------------------------------------
sub insert_res{
	my ($url, $title, $rh1, $rh2, $new) = @_;

	$$new = 0;

	#1) Lock the table (for writing)
        lock_table ("res");

	# calculate md5 for url
	my $md5 = md5_hex($url);

	#2) Check if resource exists inside the db
	$rh1->execute($md5);
	my $id = ($rh1->fetchrow_array())[0];

	#3) If the resource does not exist insert it
	if (!$id){
		print "[w] Resource $md5 does not exist: creating...\n" if $WARNINGS;
		$rh2->execute($md5, $url, $title) || die ("Error in execute");
		$id = $dbh->last_insert_id(null,null,null,null); 
		$$new=1;
	}

	#4) Unlock the table
	unlock_table("res");
	
	return $id;
}

# ---------------------------------------------------------------------------
sub insert_tagged{
	my ($uid, $tid, $rid, $date, $xh1, $xh2) = @_;

	#1) Lock the table (for writing)
        lock_table ("tagged");

	#2) Check if triple exists inside the db
	#   NOTE: triple SHOULD NOT EXIST yet!!!
	$xh1->execute($uid, $tid, $rid);
	my $id = ($xh1->fetchrow_array())[0];

	#3) If the triple does not exist insert it
	if (!$id){
		$xh2->execute($uid, $tid, $rid, $date) || die ("Error in execute");
	}

	#4) Unlock the table
	unlock_table("tagged");
}

# ---------------------------------------------------------------------------
sub getdate {
        my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime(time());
        $year+=1900; $mon=sprintf("%02d",$mon+1); $mday=sprintf("%02d",$mday);
        $hour=sprintf("%02d",$hour);
        $min=sprintf("%02d",$min);
        $sec=sprintf("%02d",$sec);
        return "$year-$mon-$mday $hour:$min:$sec";
}


# ---------------------------------------------------------------------------
# ---------------------------------------------------------------------------
#
